/*
 * This is free software; you can redistribute it and/or modify it
 * under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This code is distributed in the hope that it will be useful, but
 * WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this code; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA
 * 02111-1307, USA.  */

/* Copyright (C) 2000 Ian Zimmerman <itz@speakeasy.org> */

static char rcsid[] = "$Id: dispatch.c,v 1.28 2002/06/14 01:04:24 itz Exp $";

#ifdef HAVE_CONFIG_H
#include <config.h>
#endif

#include <ggi/gii.h>
#include <syslog.h>
#include <assert.h>
#include <string.h>
#include <errno.h>
#include <assert.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <limits.h>
#include <stdlib.h>

#include "dispatch.h"
#include "connections.h"
#include "linuxonly.h"
#include "preadwrite.h"
#include "gim_event.h"
#include "cook.h"
#include "sockets.h"
#include "exc.h"
#include "gim_common.h"
#include "options.h"

#define MAX_ARGV 20

static char reply_buf[200];
int opt_debug;

static int
translate_event (const struct gim_event_t* ev)
{
  int i;
                               /* translate event into a protocol response */
  if ((ev->kind & gim_button_ev) != 0) {
    if ((i = snprintf(reply_buf, sizeof(reply_buf), "110 %d*%x {%d,%d,%d} %dx%d %u.%06u [%x.%x]\n",
                      ev->vt, ev->kind,
                      ev->x, ev->y, ev->z,
                      ev->u.b.repeat, ev->u.b.which,
                      ev->time.tv_sec, ev->time.tv_usec,
                      ev->state, ev->modifiers)) == -1 ||
        i >= sizeof(reply_buf)) {
      syslog(LOG_ERR, "buffer too small, too many buttons?");
      exc_eraise(buffer_too_small);
    } /*if*/
  } else if ((ev->kind & gim_position_ev) != 0) {
    if ((i = snprintf(reply_buf, sizeof(reply_buf),
                      "111 %d*%x {%d(%+d),%d(%+d),%d(%+d)} |%x %u.%06u [%x.%x]\n",
                      ev->vt, ev->kind, 
                      ev->x, ev->u.d.x, ev->y, ev->u.d.y, ev->z, ev->u.d.z, 
                      ev->u.d.margin, ev->time.tv_sec, ev->time.tv_usec,
                      ev->state, ev->modifiers)) == -1 ||
        i >= sizeof(reply_buf)) {
      syslog(LOG_ERR, "buffer too small, too many pixels?");
      exc_eraise(buffer_too_small);
    } /*if*/
  } else {
    assert(0);
  } /*if*/
  return i;
}

static void
do_servinfo (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  int i;

  i = snprintf(reply_buf, sizeof(reply_buf), "210 version %s protocol %s build %s \n",
               VERSION, PROTOCOL, __DATE__);
  assert(i != -1 && i < sizeof(reply_buf));
  gim_pwrite(fd, reply_buf, i);
}

static void
do_devinfo (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  int i;
  extern gii_cmddata_getdevinfo gii_di;
  static const char gii_error_response[] = "503 gii error\n";
  
  i = snprintf(reply_buf, sizeof(reply_buf), "211 buttons %d axes %d\n",
               gii_di.num_buttons, gii_di.num_axes);
  assert(i != -1 && i < sizeof(reply_buf));
  gim_pwrite(fd, reply_buf, i);
}

static void
do_stop (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char new_error_response[] = "420 cannot stop unopened connection\n";
  static const char ok_response[] = "220 ok\n";

  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else {
    pc->state = gim_state_stopped;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_start (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char new_error_response[] = "421 cannot start unopened connection\n";
  static const char ok_response[] = "120 ok\n";

  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else {
    pc->state = gim_state_open;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_close (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char new_error_response[] = "422 cannot close unopened connection\n";
  static const char ok_response[] = "221 ok\n";

  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else {
    pc->state = gim_state_new;
    unset_connection_vt(pc);
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1); 
    syslog(LOG_INFO, "connection on tty%d closed", pc->vt);
  } /*if*/
}

static void
do_open (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char open_error_response[] = "423 connection already open\n";
  static const char ok_response[] = "121 ok\n";
  static const char auth_response[] = "530 not owner\n";
  int i;
  struct stat st;
  int vt;
  char vtname[100];

  assert(argc == 2);
  vt = atoi(argv[1]);
  if (pc->state != gim_state_new) {
    gim_pwrite(fd, open_error_response, sizeof(open_error_response)-1);    
  } else if (vt < 0 || vt > MAX_NR_CONSOLES) {
    syslog(LOG_WARNING, "open request for bogus vt %d", vt);
    exc_eraise(unrecognized_command);
  } else {
    if (vt != 0 || pc->owner != 0) {
                                /* normal user connection, check ownership */
      i = snprintf(vtname, sizeof(vtname), devfs_format, vt);
      assert(i >= 0 && i < sizeof(vtname));
      if (stat(vtname, &st) == -1) {
        i = snprintf(vtname, sizeof(vtname), nodevfs_format, vt);
        if (stat(vtname, &st) == -1) {
          int my_errno = errno;
          syslog(LOG_ERR, "stat(%s) failed: %m", vtname);
          exc_eraise2(unix_error, my_errno);
        } /*if*/
      } /*if*/
      if (pc->owner != st.st_uid) {
        syslog(LOG_WARNING, "uid %d requested open of tty%d, owned by uid %d",
               pc->owner, vt, st.st_uid);
        gim_pwrite(fd, auth_response, sizeof(auth_response)-1);
      } /*if*/
    } /*if*/
    pc->state = gim_state_stopped;
    set_connection_vt(pc, vt);
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1); 
    syslog(LOG_INFO, "connection open on tty%d by uid %d", pc->vt, pc->owner);
  } /*if*/
}

static void
do_minmod (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  unsigned minmod;
  static const char new_error_response[] = "424 cannot set mask of unopened connection\n";
  static const char open_error_repsonse[] = "425 cannot set mask of active connection\n";
  static const char ok_response[] = "222 ok\n";

  if (sscanf(argv[1], "%x", &minmod) != 1) {
    syslog(LOG_WARNING, "invalid modifier mask: %s", argv[1]);
    exc_eraise(unrecognized_command);
  } /*if*/
  
  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else if (pc->state == gim_state_open) {
    gim_pwrite(fd, open_error_repsonse, sizeof(open_error_repsonse)-1);
  } else {
    assert(pc->state == gim_state_stopped);
    pc->min_mod = minmod;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_maxmod (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  unsigned maxmod;
  static const char new_error_response[] = "426 cannot set mask of unopened connection\n";
  static const char open_error_repsonse[] = "427 cannot set mask of active connection\n";
  static const char ok_response[] = "223 ok\n";

  if (sscanf(argv[1], "%x", &maxmod) != 1) {
    syslog(LOG_WARNING, "invalid modifier mask: %s", argv[1]);
    exc_eraise(unrecognized_command);
  } /*if*/
  
  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else if (pc->state == gim_state_open) {
    gim_pwrite(fd, open_error_repsonse, sizeof(open_error_repsonse)-1);
  } else {
    assert(pc->state == gim_state_stopped);
    pc->max_mod = maxmod;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_handle (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  unsigned handle;
  static const char new_error_response[] = "426 cannot set mask of unopened connection\n";
  static const char open_error_repsonse[] = "427 cannot set mask of active connection\n";
  static const char ok_response[] = "223 ok\n";

  if (sscanf(argv[1], "%x", &handle) != 1) {
    syslog(LOG_WARNING, "invalid event mask: %s", argv[1]);
    exc_eraise(unrecognized_command);
  } /*if*/
  
  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else if (pc->state == gim_state_open) {
    gim_pwrite(fd, open_error_repsonse, sizeof(open_error_repsonse)-1);
  } else {
    assert(pc->state == gim_state_stopped);
    pc->handle = handle;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_pass (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  unsigned pass;
  static const char new_error_response[] = "426 cannot set mask of unopened connection\n";
  static const char open_error_repsonse[] = "427 cannot set mask of active connection\n";
  static const char ok_response[] = "223 ok\n";

  if (sscanf(argv[1], "%x", &pass) != 1) {
    syslog(LOG_WARNING, "invalid event mask: %s", argv[1]);
    exc_eraise(unrecognized_command);
  } /*if*/
  
  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else if (pc->state == gim_state_open) {
    gim_pwrite(fd, open_error_repsonse, sizeof(open_error_repsonse)-1);
  } else {
    assert(pc->state == gim_state_stopped);
    pc->pass = pass;
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1);
  } /*if*/
}

static void
do_getmasks (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  int i;
  static const char new_error_response[] = "428 cannot get mask of unopened connection\n";

  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else {
    i = snprintf(reply_buf, sizeof(reply_buf), "224 min 0x%x max 0x%x handle 0x%x pass 0x%x\n",
                 pc->min_mod, pc->max_mod, pc->handle, pc->pass);
    assert(i != -1 && i < sizeof(reply_buf));
    gim_pwrite(fd, reply_buf, i);
  } /*if*/
}

static void
do_drawptr(int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char new_error_response[] = "422 cannot draw pointer on unopened connection\n";
  static const char open_error_repsonse[] = "427 cannot set mask of active connection\n";
  static const char ok_response[] = "225 ok\n";
  int x, y;

  assert(argc == 3);
  if (sscanf(argv[1],"%d", &x) != 1 || sscanf(argv[2],"%d", &y) != 1) {
    syslog(LOG_WARNING, "bad pointer coordinates: %s %s", argv[1], argv[2]);
    exc_eraise(unrecognized_command);
  } /*if*/
  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else if (pc->state == gim_state_open) {
    gim_pwrite(fd, open_error_repsonse, sizeof(open_error_repsonse)-1);
  } else {
    draw_pointer(x, y, pc->vt);
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1); 
  } /*if*/
}

static void
do_disable_paste (int fd, struct gim_connect_t* pc, int argc, char** argv)
{
  static const char new_error_response[] = "422 cannot disable paste on unopened connection\n";
  static const char ok_response[] = "226 ok\n";

  if (pc->state == gim_state_new) {
    gim_pwrite(fd, new_error_response, sizeof(new_error_response)-1);
  } else {
    disable_paste(pc->vt);
    gim_pwrite(fd, ok_response, sizeof(ok_response)-1); 
  } /*if*/
}

static void
dispatch_selection (const struct gim_event_t* pgev)
{
  static int xs, ys;

  switch (pgev->kind & (gim_raw_ev | gim_multi_ev)) {
  case gim_move_ev:
    draw_pointer (pgev->x, pgev->y, pgev->vt);
    break;

  case (gim_drag_ev | gim_single_ev):
    if (0 != (pgev->state & 1)) copy_selection (xs, ys, pgev->x, pgev->y, 1, pgev->vt);
    if (opt_pointer) draw_pointer (pgev->x, pgev->y, pgev->vt);
    break;

  case (gim_drag_ev | gim_double_ev):
    if (0 != (pgev->state & 1)) copy_selection (xs, ys, pgev->x, pgev->y, 2, pgev->vt);
    if (opt_pointer) draw_pointer (pgev->x, pgev->y, pgev->vt);
    break;

  case (gim_drag_ev | gim_triple_ev):
    if (0 != (pgev->state & 1)) copy_selection (xs, ys, pgev->x, pgev->y, 3, pgev->vt);
    if (opt_pointer) draw_pointer (pgev->x, pgev->y, pgev->vt);
    break;

  case (gim_down_ev | gim_single_ev):
    switch (pgev->u.b.which) {
    case 1:                     /* left */
      xs = pgev->x; ys = pgev->y;
      copy_selection (xs, ys, pgev->x, pgev->y, 1, pgev->vt);
      break;
    case 2:                     /* right */
      copy_selection (xs, ys, pgev->x, pgev->y, 1, pgev->vt);
      break;
    case 3:                     /* middle */
      paste_selection (pgev->vt);
      break;
    } /*switch*/
    break;

  case (gim_down_ev | gim_double_ev):
    switch (pgev->u.b.which) {
    case 1:                     /* left */
      xs = pgev->x; ys = pgev->y;
      copy_selection (xs, ys, pgev->x, pgev->y, 2, pgev->vt);
      break;
    case 2:                     /* right */
      copy_selection (xs, ys, pgev->x, pgev->y, 2, pgev->vt);
      break;
    } /*switch*/
    break;

  case (gim_down_ev | gim_triple_ev):
    switch (pgev->u.b.which) {
    case 1:                     /* left */
      xs = pgev->x; ys = pgev->y;
      copy_selection (xs, ys, pgev->x, pgev->y, 3, pgev->vt);
      break;
    case 2:                     /* right */
      copy_selection (xs, ys, pgev->x, pgev->y, 3, pgev->vt);
      break;
    } /*switch*/
    break;

  } /*switch*/

}

void
dispatch_event (const gii_event* ev)
{
  int len;
  struct gim_connect_t* pc;
  struct gim_event_t gev;
  int pass;

  /* cook_event returns 0 if event scales down to nothing */
  if (!cook_event(ev, &gev)) return;
  len = translate_event(&gev);
  
  pass = 1;
  pc = get_active_connection(gev.vt, gev.modifiers);
  if (pc != 0) {
    if ((gev.kind & pc->handle) != 0) gim_pwrite(pc->fd, reply_buf, len);
    pass = ((gev.kind & pc->pass) != 0);
  } /*if*/
  if (gev.vt != 0 && pass) {
    pc = get_active_connection(0, gev.modifiers);
    if (pc != 0) {
      if ((gev.kind & pc->handle) != 0) gim_pwrite(pc->fd, reply_buf, len);
      pass = ((gev.kind & pc->pass) != 0);      
    } /*if*/
  } /*if*/

  if (opt_debug) syslog(LOG_DEBUG, reply_buf);

  /* if nobody was interested, at least do selection */
  if (pass) dispatch_selection(&gev);
}

#define TABLE_SIZE(t) (sizeof(t)/sizeof((t)[0]))

void
handle_client_request (int fd, int* max_client, fd_set* clients)
{
  struct gim_connect_t* pc;
  char cmdbuf[UCHAR_MAX];
  ssize_t len;
  char* p;
  int argc;
  char* argv[MAX_ARGV];
  static const char unrecognized_response[] = "500 unrecognized command\n";
  static const char invalid_response[] = "501 invalid argument\n";
  typedef void (*hfn)(int, struct gim_connect_t*, int, char**);
  struct hfn_table_entry { char cmd_name[20]; int argc; hfn handler; };
  struct hfn_table_entry* p_entry;

  static struct hfn_table_entry cmd_table[] = {
    { "close", 1, do_close },
    { "devinfo", 1, do_devinfo },
    { "disable_paste", 1, do_disable_paste },
    { "drawptr", 3, do_drawptr },
    { "getmasks", 1, do_getmasks },
    { "handle", 2, do_handle },
    { "maxmod", 2, do_maxmod },
    { "minmod", 2, do_minmod },
    { "open", 2, do_open },
    { "pass", 2, do_pass },
    { "servinfo", 1, do_servinfo },
    { "start", 1, do_start },
    { "stop", 1, do_stop }
  };

  pc = get_connection_by_fd(fd);
  assert(pc);
  memset(cmdbuf, 0, sizeof(cmdbuf));

  exc_try {
    len = gim_pread(fd, cmdbuf, sizeof(cmdbuf));
    if (len == 0 || cmdbuf[len-1] != '\n') {
      syslog(LOG_WARNING, "%s (len = %d): client command not terminated by newline, vt = %d", cmdbuf, len, pc->vt);
      exc_eraise(malformed_command);
    } /*if*/
    cmdbuf[len-1] = '\0';
    p = strtok(cmdbuf, " \t\r\n");
    argc = 0;
    while (p != 0 && argc < MAX_ARGV) {
      argv[argc] = p;
      p = strtok(0, " \t\r\n");
      argc = argc + 1;
    } /*while*/
    if (argc > MAX_ARGV-1) {
      syslog(LOG_WARNING, "too many tokens in client command, vt = %d", pc->vt);
      exc_eraise(unrecognized_command);
    } /*if*/
    if (argc == 0) {
      syslog(LOG_WARNING, "empty client command, vt = %d", pc->vt);
      exc_eraise(unrecognized_command);
    } /*if*/
    argv[argc] = 0;
    p_entry = bsearch(argv[0], cmd_table, TABLE_SIZE(cmd_table),
                      sizeof(struct hfn_table_entry), (comparison_fn_t)strcmp);
    if (p_entry == 0) {
      syslog(LOG_WARNING, "unrecognized command: %s, vt = %d", argv[0], pc->vt);
      exc_eraise(unrecognized_command);
    } /*if*/
    if (p_entry->argc != argc) {
      syslog(LOG_WARNING, "wrong number of arguments: %s, vt = %d", argv[0], pc->vt);
      exc_eraise(unrecognized_command);
    } /*if*/

    (*p_entry->handler)(fd, pc, argc, argv);

  }
  exc_except(unrecognized_command) {
    gim_pwrite(fd, unrecognized_response, sizeof(unrecognized_response)-1);
    if (pc->state != gim_state_new) {
      syslog(LOG_DEBUG, "%s: %d: unrecognized command (on tty%d)", exc_file, exc_line, pc->vt);
    } /*if*/
    syslog(LOG_DEBUG, "handling client, continuing");
  }
  exc_except(exc_any) {
    if (pc->state != gim_state_new) {
      syslog(LOG_DEBUG, "%s: %d: %s (on tty%d)", exc_file, exc_line, exc_name, pc->vt);
      if (exc_info_size >= sizeof(unsigned)) {
        syslog(LOG_DEBUG, "info starts: %x", *(unsigned *)exc_info);
      } /*if*/
    } /*if*/
    syslog(LOG_DEBUG, "handling client, closing and continuing");
    unset_connection_vt(pc);
    delete_connection(pc);
    delete_client(fd, max_client, clients);
  } exc_end;

}
